import { Permissions } from './../generated/permissions.js';

export class PermissionsManager {
    constructor() {
        this.hasPermissions = true;
        this.global = Permissions.globalPermissions || {};
        this.elementsPermissions = Permissions.elementsPermissions || {};
        this.objectsPermissions = Permissions.objectsPermissions || {};
        this.windowsPermissions = Permissions.windowsPermissions || {};
        this.operatorsRoles = Permissions.operatorsRoles || {};
        this.operatorsMnemos = Permissions.operatorsMnemos || {};
        this.logActionTry = Permissions.logActionTry || {};
        this.emptyPermConst = "0000";
        this.allRoleConst = "_All";
        this.permissionTypes = {
            Action: "Действие",
            Control: "Управление",
            OpenWindow: "Открытие окна",
            ChangeUsers: "Смена пользователя",
            ChangeSelfPassword: "Изменение пароля",
            JournalAck: "Квитирование сообщения",
            JournalGroupAck: "Групповое квитирование сообщений",
            JournalUseFilters: "Использование фильтров в журнале",
            JournalPrint: "Печать сообщений",
            JournalSave: "Экспорт сообщений",
            TrendAddPens: "Добавление перьев",
            TrendDeletePens: "Удаление перьев",
            TrendChangePens: "Изменение настройки перьев, изменение шкалы",
            TrendSave: "Экспорт графика",
            TrendPrint: "Печать графика",
            DataTableEdit: "Редактирование ячеек таблицы",
            DataTableFilter: "Использование фильтров в таблице",
        }
    }

    openUserWindow(customWin) {
        if (this.hasPermissions) {
            const permit = this.checkOpenWindowRights(customWin);
            //                                                /\
            //стартовые окна не могут быть экземплярами, поэтому контекст не обязателен
            if (!permit) {
                const item = {
                    type: 'error',
                    time: new Date().toLocaleString(),
                    title: 'Нет прав на заданное окно', text: 'Нет прав на открытие заданного стартового окна. Проверьте настройки прав для данного пользователя'
                };
                $ns.add(item);
                return false;
            }
        }
        const frame = document.getElementById($WinDefs.constructor.mainFrame);
        frame.openWindow($WinDefs.winDef[customWin]);
        return true;
    }

    checkCurrentWindow() {
        const firstWin = document.getElementById($WinDefs.constructor.mainFrame).firstChild;
        if (firstWin) {
            const curWin = firstWin.id;
            return this.checkOpenWindowRights(curWin, firstWin);
        }
    }

    getStartWindow() {
        if (Object.keys(this.operatorsMnemos).length > 0) {
            let intersec = this.getUsersIntersec(this.operatorsMnemos);
            if (intersec && intersec.length === 1) return this.operatorsMnemos[intersec[0]];
            if (intersec && intersec.length > 1) {
                //если пользователь состоит в нескольких группах 
                //и для этих групп заданы стартовые окна
                //вернем первое разрешенное окно из списка групп
                for (let i = 0; i < intersec.length; i++) {
                    const startMnemo = this.operatorsMnemos[intersec[i]];
                    if (this.checkOpenWindowRights(startMnemo)) {
                        return startMnemo;
                    }
                }
            }
        }
    }

    //for silent mode, where we dont have confirmation and action based opening
    checkOpenWindowRights(winId, context) {
        const entry = this.getEntryFromWindowType(winId, "OpenWindow", context);
        return this.checkAllow(entry);
    }

    setOperatorRoles(operator, roles) {
        this.currentOp = operator;
        if (typeof (roles) !== 'undefined' && roles.length > 0) {
            if (typeof roles[0] === 'string') {
                this.operatorsRoles[operator] = roles;
            } else if (roles[0] !== null && typeof roles[0] === 'object') {
                this.operatorsRoles[operator] = roles.map(x => x.name);
            }
        }
        this.changeUsersList(operator);
    }

    changeUsersList(operator) {
        if (!operator) {
            operator = this.currentOp;
        }
        let roles = this.operatorsRoles && new Set(this.operatorsRoles[operator]);
        if (typeof (roles) !== 'undefined' && roles.size > 0) {
            roles.add(operator);
            roles.add(this.allRoleConst);
            this.usersList = Array.from(roles);
        } else {
            this.usersList = [operator, this.allRoleConst];
        }
    }

    getParentWindow(element) {
        let temp = element;
        while (temp && temp.nodeName !== "MS-WINDOW") {
            temp = temp.parentNode;
        }
        return temp;
    }

    getObjectId(typeId, context) {
        let objId = Number($WinDefs.winDef[typeId] && $WinDefs.winDef[typeId].objectid);
        if (!objId && context) {
            objId = $sw.ItemSubscription._getParentParam(context, 'objectid');
        }
        return objId;
    }

    controlEnable(childs) {
        childs && childs.forEach(ch => {
            if (typeof ch.setPermission === 'function') {
                if (ch.childNodes.length > 0) { //dont disable containers, 
                    //                          it possibly contains special allow
                    this.controlEnable(ch.childNodes);
                } else {
                    let entry = this.getEntryFromElement(ch, "Control");
                    let disable = !this.checkAllow(entry);
                    ch.setControlDisable(disable);
                }
            }
        });
    }

    checkEnable(element) {
        return this.checkAllow(this.getEntryFromElement(element, "Control"));
    }

    getEntryFromElement(element, permType) {
        let result;
        if (result = element.getPermission(this.usersList, permType)) {
            return result;
        }
        const parentWin = this.getParentWindow(element);
        const windowTypeId = parentWin && parentWin.typeid;
        const windowId = parentWin && parentWin.id;
        const permContainer = this.elementsPermissions[windowTypeId];

        if (typeof (permContainer) !== 'undefined') {
            let node = element;
            do {
                const elPerm = permContainer[node.id];
                const users = typeof (elPerm) !== 'undefined'
                    && this.getUsersIntersec(elPerm[permType]);
                if (users && users.length > 0) {
                    result = this.getReducedEntry(elPerm[permType], users);
                    break;
                }
                if (node.id == windowId) break;
                node = node.parentNode;
            } while (node.id != windowId);
        }

        if (!result) {
            result = this.getEntryFromWindowType(windowTypeId, permType, element);
        }
        element.setPermission(this.usersList, permType, result);
        return result;
    }

    getEntryFromWindowType(typeId, permType, context, objectId) {
        const winPerms = this.windowsPermissions[typeId];
        if (typeof (winPerms) !== 'undefined') {
            const perms = winPerms[permType];
            const users = this.getUsersIntersec(perms);
            if (users && users.length > 0) {
                return this.getReducedEntry(perms, users);
            }
        }
        if (!objectId) {
            objectId = this.getObjectId(typeId, context);
        }
        return this.getEntryFromObject(objectId, permType);
    }

    getEntryFromObject(objectId, permType) {
        const objPerm = this.objectsPermissions[objectId];
        if (typeof (objPerm) !== 'undefined') {
            const perms = objPerm[permType];
            const objParent = objPerm.parentId;
            const users = this.getUsersIntersec(perms);
            if (users && users.length > 0) {
                return this.getReducedEntry(perms, users);
            } else if (typeof (this.objectsPermissions[objParent]) !== 'undefined') {
                return this.getEntryFromObject(objParent, permType);
            }
        }
        const globalPerm = this.global[permType];
        if (typeof (globalPerm) !== 'undefined') {
            return this.getReducedEntry(this.global[permType], this.usersList);
        } else return this.emptyPermConst;

    }

    /*from perms list exract users intersected with current users list */
    getUsersIntersec(perms) {
        return typeof (perms) !== 'undefined'
            && Object.keys(perms).filter(x => this.usersList.includes(x));
    }

    /*'users' - list of the intersections with defined in 'rights' operators */
    getReducedEntry(rights, users) {
        if (users.length === 1) return rights[users[0]];
        else {
            let arr = users.reduce((acc, cur) => {
                if (rights[cur])
                    acc.push(rights[cur]);
                return acc;
            }, []);
            return this.reduceEntries(...arr);
        }
    }

    reduceEntries(...entries) {
        let reducedEntry = entries.reduce((acc, cur) => acc | parseInt(cur, 2), 0).toString(2);
        /*adding leading zeros for normalize length*/
        return "0000".substr(reducedEntry.length) + reducedEntry;
    }

    resolveOpenWindow(context, windId, objectId, callback) {
        const optype = "OpenWindow";
        const entryForWindow = this.getEntryFromWindowType(windId, optype, context, objectId);
        const entryForContext = this.getEntryFromElement(context, optype);
        const result = this.reduceEntries(entryForContext, entryForWindow);
        this.resolveAction(context, result, optype, callback);
    }

    resolveControl(context, optypeDesc, callback, options) {
        const optype = optypeDesc === "Action" ? "Control" : optypeDesc;
        let entry = this.getEntryFromElement(context, optype);
        this.resolveAction(context, entry, optype, callback, options);
    }

    resolveAction(context, entry, optype, callback, options) {
        const { customText, customDialogText, uid = 0, cancel, forced = false } = options || {};
        if (this.checkAllow(entry)) {
            if (this.Confirmation(entry) && !forced) {
                if (this.getLogActionTryFlag() && this.WriteLog(entry)) {
                    this.reportAction(context, optype, customText, true);
                }
                return $confirm(
                    this.executeOperation.bind(this, context, entry, optype, callback, customText),
                    customDialogText || this.getPermissionText(optype),
                    uid,
                    cancel);
            } else {
                return this.executeOperation(context, entry, optype, callback, customText);
            }
        }
    }

    executeOperation(context, entry, optype, callback, customText) {
        if (this.WriteLog(entry)) {
            this.reportAction(context, optype, customText);
        }
        return callback && callback();
    }

    getLogActionTryFlag() {
        let intersec = this.getUsersIntersec(this.logActionTry);
        if (intersec.length === 1) return this.logActionTry[intersec[0]];
        else {
            //если пользователь состоит в нескольких группах 
            //и флаг логировать попытки действий true хотя бы в одной, то логируем
            return intersec.reduce((acc, cur) => this.logActionTry[cur] || acc, false);
        }
    }

    reportAction(context, optype, customText, logActionTry) {
        const wind = this.getParentWindow(context);
        const windId = wind && wind.typeid;
        const objId = this.getObjectId(windId, context);
        if (typeof customText === "undefined") {
            const permname = this.getPermissionText(optype);
            customText = `${permname}: ${context.elname}`;
        }
        if (logActionTry) {
            customText = "Попытка действия: " + customText;
        }

        $sw.ReportAction(context.id, objId, customText, context.elname);
    }


    getPermissionText(optype) {
        const result = this.permissionTypes[optype];
        if (typeof result === "undefined") return optype;
        return result;
    }

    checkAllow(entry) {
        return this.Enable(entry) && !this.Disable(entry);
    }

    Enable(entry) {
        return !!(entry && entry[0] === "1");
    }

    Disable(entry) {
        return !!(entry && entry[1] === "1");
    }

    Confirmation(entry) {
        return !!(entry && entry[2] === "1");
    }

    WriteLog(entry) {
        return !!(entry && entry[3] === "1");
    }
}